@@ -37,10 +37,8 @@ module Data.Csv.Encoding
     , recordSep
     ) where
 
-import Blaze.ByteString.Builder (Builder, fromByteString, fromWord8,
-                                 toLazyByteString, toByteString)
-import Blaze.ByteString.Builder.Char8 (fromString)
-import Control.Applicative as AP (Applicative(..), (<|>), optional)
+import Data.ByteString.Builder
+import Control.Applicative as AP (Applicative(..), (<|>))
 import Data.Attoparsec.ByteString.Char8 (endOfInput)
 import qualified Data.Attoparsec.ByteString.Lazy as AL
 import qualified Data.ByteString as B
@@ -62,7 +60,7 @@ import Data.Csv.Parser hiding (csv, csvWithHeader)
 import qualified Data.Csv.Parser as Parser
 import Data.Csv.Types hiding (toNamedRecord)
 import qualified Data.Csv.Types as Types
-import Data.Csv.Util (blankLine, endOfLine)
+import Data.Csv.Util (blankLine, endOfLine, toStrict)
 
 
 -- TODO: 'encode' isn't as efficient as it could be.
@@ -232,8 +230,8 @@ encodeOptionsError = error $ "Data.Csv: " ++
 -- | Encode a single record, without the trailing record separator
 -- (i.e. newline).
 encodeRecord :: Quoting -> Word8 -> Record -> Builder
-encodeRecord qtng delim = mconcat . intersperse (fromWord8 delim)
-                     . map fromByteString . map (escape qtng delim) . V.toList
+encodeRecord qtng delim = mconcat . intersperse (word8 delim)
+                     . map byteString . map (escape qtng delim) . V.toList
 {-# INLINE encodeRecord #-}
 
 -- | Encode a single named record, without the trailing record
@@ -246,23 +244,22 @@ encodeNamedRecord hdr qtng delim =
 escape :: Quoting -> Word8 -> B.ByteString -> B.ByteString
 escape !qtng !delim !s
     | (qtng == QuoteMinimal &&
-        B.any (\ b -> b == dquote || b == delim || b == nl || b == cr || b == sp) s
+        B.any (\ b -> b == dquote || b == delim || b == nl || b == cr) s
       ) || qtng == QuoteAll
-         = toByteString $
-            fromWord8 dquote
+         = toStrict . toLazyByteString $
+            word8 dquote
             <> B.foldl
                 (\ acc b -> acc <> if b == dquote
-                    then fromByteString "\"\""
-                    else fromWord8 b)
+                    then byteString "\"\""
+                    else word8 b)
                 mempty
                 s
-            <> fromWord8 dquote
+            <> word8 dquote
     | otherwise = s
   where
     dquote = 34
     nl     = 10
     cr     = 13
-    sp     = 32
 
 -- | Like 'encodeByName', but lets you customize how the CSV data is
 -- encoded.
@@ -316,8 +313,8 @@ moduleError func msg = error $ "Data.Csv.Encoding." ++ func ++ ": " ++ msg
 {-# NOINLINE moduleError #-}
 
 recordSep :: Bool -> Builder
-recordSep False = fromWord8 10 -- new line (\n)
-recordSep True  = fromString "\r\n"
+recordSep False = word8 10 -- new line (\n)
+recordSep True  = string8 "\r\n"
 
 unlines :: Builder -> [Builder] -> Builder
 unlines _ [] = mempty
@@ -357,18 +354,16 @@ decodeWithP p s =
 csv :: FromRecord a => DecodeOptions -> AL.Parser (V.Vector a)
 csv !opts = do
     vals <- records
-    _ <- optional endOfLine
-    endOfInput
     return $! V.fromList vals
   where
     records = do
         !r <- record (decDelimiter opts)
         if blankLine r
-            then (endOfLine *> records) <|> pure []
+            then (endOfInput *> pure []) <|> (endOfLine *> records)
             else case runParser (parseRecord r) of
                 Left msg  -> fail $ "conversion error: " ++ msg
                 Right val -> do
-                    !vals <- (endOfLine *> records) <|> AP.pure []
+                    !vals <- (endOfInput *> AP.pure []) <|> (endOfLine *> records)
                     return (val : vals)
 {-# INLINE csv #-}
 
@@ -378,19 +373,17 @@ csvWithHeader :: FromNamedRecord a => DecodeOptions
 csvWithHeader !opts = do
     !hdr <- header (decDelimiter opts)
     vals <- records hdr
-    _ <- optional endOfLine
-    endOfInput
     let !v = V.fromList vals
     return (hdr, v)
   where
     records hdr = do
         !r <- record (decDelimiter opts)
         if blankLine r
-            then (endOfLine *> records hdr) <|> pure []
+            then (endOfInput *> pure []) <|> (endOfLine *> records hdr)
             else case runParser (convert hdr r) of
                 Left msg  -> fail $ "conversion error: " ++ msg
                 Right val -> do
-                    !vals <- (endOfLine *> records hdr) <|> pure []
+                    !vals <- (endOfInput *> pure []) <|> (endOfLine *> records hdr)
                     return (val : vals)
 
     convert hdr = parseNamedRecord . Types.toNamedRecord hdr
