@@ -8,6 +8,7 @@ module Data.Csv.Util
     , doubleQuote
     , newline
     , cr
+    , toStrict
     ) where
 
 import Control.Applicative ((<|>))
@@ -22,6 +23,15 @@ import Data.Attoparsec.ByteString (Parser)
 import Control.Applicative ((*>))
 #endif
 
+#if MIN_VERSION_bytestring(0,10,0)
+import Data.ByteString.Lazy (toStrict)
+#else
+import qualified Data.ByteString.Lazy as L
+
+toStrict :: L.ByteString -> B.ByteString
+toStrict = B.concat . L.toChunks
+#endif
+
 -- | A strict version of 'Data.Functor.<$>' for monads.
 (<$!>) :: Monad m => (a -> b) -> m a -> m b
 f <$!> m = do
